import re
from urllib.parse import parse_qs, urlencode

from benedict.serializers.abstract import AbstractSerializer


class QueryStringSerializer(AbstractSerializer):
    """
    This class describes a query-string serializer.
    """

    def __init__(self):
        super().__init__(
            extensions=[
                "qs",
                "querystring",
            ],
        )

    def decode(self, s, **kwargs):
        flat = kwargs.pop("flat", True)
        qs_re = r"(?:([\w\-\%\+\.\|]+\=[\w\-\%\+\.\|]*)+(?:[\&]{1})?)+"
        qs_pattern = re.compile(qs_re)
        if qs_pattern.match(s):
            data = parse_qs(s)
            if flat:
                data = {key: value[0] for key, value in data.items()}
            return data
        raise ValueError(f"Invalid query string: {s}")

    def encode(self, d, **kwargs):
        data = urlencode(d, **kwargs)
        return data



from benedict.dicts.base import BaseDict
from benedict.dicts.keylist import keylist_util
from benedict.utils import type_util


class KeylistDict(BaseDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __contains__(self, key):
        if type_util.is_list_or_tuple(key):
            return self._contains_by_keys(key)
        return super().__contains__(key)

    def _contains_by_keys(self, keys):
        parent, _, _ = keylist_util.get_item(self, keys)
        if type_util.is_dict_or_list_or_tuple(parent):
            return True
        return False

    def __delitem__(self, key):
        if type_util.is_list_or_tuple(key):
            self._delitem_by_keys(key)
            return
        super().__delitem__(key)

    def _delitem_by_keys(self, keys):
        parent, key, _ = keylist_util.get_item(self, keys)
        if type_util.is_dict_or_list(parent):
            del parent[key]
            return
        elif type_util.is_tuple(parent):
            # raise the standard TypeError
            del parent[key]
        raise KeyError(f"Invalid keys: {keys!r}")

    def __getitem__(self, key):
        if type_util.is_list_or_tuple(key):
            return self._getitem_by_keys(key)
        return super().__getitem__(key)

    def _getitem_by_keys(self, keys):
        parent, key, _ = keylist_util.get_item(self, keys)
        if type_util.is_dict_or_list_or_tuple(parent):
            return parent[key]
        raise KeyError(f"Invalid keys: {keys!r}")

    def __setitem__(self, key, value):
        if type_util.is_list_or_tuple(key):
            self._setitem_by_keys(key, value)
            return
        super().__setitem__(key, value)

    def _setitem_by_keys(self, keys, value):
        keylist_util.set_item(self, keys, value)

    def get(self, key, default=None):
        if type_util.is_list_or_tuple(key):
            return self._get_by_keys(key, default)
        return super().get(key, default)

    def _get_by_keys(self, keys, default=None):
        parent, key, _ = keylist_util.get_item(self, keys)
        if type_util.is_dict(parent):
            return parent.get(key, default)
        elif type_util.is_list_or_tuple(parent):
            return parent[key]
        return default

    def pop(self, key, *args):
        if type_util.is_list_or_tuple(key):
            return self._pop_by_keys(key, *args)
        return super().pop(key, *args)

    def _pop_by_keys(self, keys, *args):
        parent, key, _ = keylist_util.get_item(self, keys)
        if type_util.is_dict(parent):
            return parent.pop(key, *args)
        elif type_util.is_list(parent):
            return parent.pop(key)
        elif type_util.is_tuple(parent):
            # raise the standard TypeError
            del parent[key]
        if args:
            return args[0]
        raise KeyError(f"Invalid keys: {keys!r}")

    def set(self, key, value):
        self[key] = value

    def setdefault(self, key, default=None):
        if key not in self:
            self[key] = default
            return default
        return self[key]



from configparser import DEFAULTSECT as default_section
from configparser import ConfigParser
from io import StringIO

from benedict.serializers.abstract import AbstractSerializer
from benedict.utils import type_util


class INISerializer(AbstractSerializer):
    """
    This class describes an ini serializer.
    """

    def __init__(self):
        super().__init__(
            extensions=[
                "ini",
            ],
        )

    @staticmethod
    def _get_section_option_value(parser, section, option):
        value = None
        funcs = [parser.getint, parser.getfloat, parser.getboolean, parser.get]
        for func in funcs:
            try:
                value = func(section, option)
                break
            except ValueError:
                continue
        return value

    def decode(self, s, **kwargs):
        parser = ConfigParser(**kwargs)
        parser.read_string(s)
        data = {}
        for option, _ in parser.defaults().items():
            data[option] = self._get_section_option_value(
                parser, default_section, option
            )
        for section in parser.sections():
            data[section] = {}
            for option, _ in parser.items(section):
                data[section][option] = self._get_section_option_value(
                    parser, section, option
                )
        return data

    def encode(self, d, **kwargs):
        parser = ConfigParser(**kwargs)
        for key, value in d.items():
            if not type_util.is_dict(value):
                parser.set(default_section, key, f"{value}")
                continue
            section = key
            parser.add_section(section)
            for option_key, option_value in value.items():
                parser.set(section, option_key, f"{option_value}")
        str_data = StringIO()
        parser.write(str_data)
        return str_data.getvalue()



from benedict.dicts import KeylistDict
from benedict.dicts.keypath import keypath_util


class KeypathDict(KeylistDict):
    _keypath_separator = None

    def __init__(self, *args, **kwargs):
        self._keypath_separator = kwargs.pop("keypath_separator", ".")
        check_keys = kwargs.pop("check_keys", True)
        super().__init__(*args, **kwargs)
        if check_keys:
            keypath_util.check_keys(self, self._keypath_separator)

    @property
    def keypath_separator(self):
        return self._keypath_separator

    @keypath_separator.setter
    def keypath_separator(self, value):
        keypath_util.check_keys(self, value)
        self._keypath_separator = value

    def __contains__(self, key):
        return super().__contains__(self._parse_key(key))

    def __delitem__(self, key):
        super().__delitem__(self._parse_key(key))

    def __getitem__(self, key):
        return super().__getitem__(self._parse_key(key))

    def __setitem__(self, key, value):
        keypath_util.check_keys(value, self._keypath_separator)
        super().__setitem__(self._parse_key(key), value)

    def __setstate__(self, state):
        super().__setstate__(state)
        self._keypath_separator = state["_keypath_separator"]

    def _parse_key(self, key):
        keys = keypath_util.parse_keys(key, self._keypath_separator)
        keys_count = len(keys)
        if keys_count == 0:
            return None
        elif keys_count == 1:
            return keys[0]
        return keys

    @classmethod
    def fromkeys(cls, sequence, value=None):
        d = cls()
        for key in sequence:
            d[key] = value
        return d

    def get(self, key, default=None):
        return super().get(self._parse_key(key), default)

    def pop(self, key, *args):
        return super().pop(self._parse_key(key), *args)

    def update(self, other):
        keypath_util.check_keys(other, self._keypath_separator)
        super().update(other)



from benedict.core import clone as _clone


class BaseDict(dict):
    _dict = None
    _pointer = False

    @classmethod
    def _get_dict_or_value(cls, value):
        value = value.dict() if isinstance(value, cls) else value
        if isinstance(value, dict):
            for key in value.keys():
                key_val = value[key]
                if isinstance(key_val, cls):
                    key_val = cls._get_dict_or_value(value[key])
                    value[key] = key_val
        return value

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            self._dict = self._get_dict_or_value(args[0])
            self._pointer = True
            super().__init__(self._dict)
            return
        self._dict = None
        self._pointer = False
        super().__init__(*args, **kwargs)

    def __bool__(self):
        if self._pointer:
            return bool(self._dict)
        return len(self.keys()) > 0

    def __contains__(self, key):
        if self._pointer:
            return key in self._dict
        return super().__contains__(key)

    def __deepcopy__(self, memo):
        obj = self.__class__()
        for key, value in self.items():
            obj[key] = _clone(value, memo=memo)
        return obj

    def __delitem__(self, key):
        if self._pointer:
            del self._dict[key]
            return
        super().__delitem__(key)

    def __eq__(self, other):
        if self._pointer:
            return self._dict == other
        return super().__eq__(other)

    def __getitem__(self, key):
        if self._pointer:
            return self._dict[key]
        return super().__getitem__(key)

    def __iter__(self):
        if self._pointer:
            return iter(self._dict)
        return super().__iter__()

    def __len__(self):
        if self._pointer:
            return len(self._dict)
        return super().__len__()

    def __repr__(self):
        if self._pointer:
            return repr(self._dict)
        return super().__repr__()

    def __setitem__(self, key, value):
        value = self._get_dict_or_value(value)
        if self._pointer:
            is_dict_item = key in self._dict and isinstance(self._dict[key], dict)
            is_dict_value = isinstance(value, dict)
            if is_dict_item and is_dict_value:
                self._dict[key].clear()
                self._dict[key].update(value)
                return
            self._dict[key] = value
            return
        super().__setitem__(key, value)

    def __setstate__(self, state):
        self._dict = state["_dict"]
        self._pointer = state["_pointer"]

    def __str__(self):
        if self._pointer:
            return str(self._dict)
        return super().__str__()

    def clear(self):
        if self._pointer:
            self._dict.clear()
            return
        super().clear()

    def copy(self):
        if self._pointer:
            return self._dict.copy()
        return super().copy()

    def dict(self):
        if self._pointer:
            return self._dict
        return self

    def get(self, key, default=None):
        if self._pointer:
            return self._dict.get(key, default)
        return super().get(key, default)

    def items(self):
        if self._pointer:
            return self._dict.items()
        return super().items()

    def keys(self):
        if self._pointer:
            return self._dict.keys()
        return super().keys()

    def pop(self, key, *args):
        if self._pointer:
            return self._dict.pop(key, *args)
        return super().pop(key, *args)

    def setdefault(self, key, default=None):
        default = self._get_dict_or_value(default)
        if self._pointer:
            return self._dict.setdefault(key, default)
        return super().setdefault(key, default)

    def update(self, other):
        other = self._get_dict_or_value(other)
        if self._pointer:
            self._dict.update(other)
            return
        super().update(other)

    def values(self):
        if self._pointer:
            return self._dict.values()
        return super().values()



from benedict.exceptions import ExtrasRequireModuleNotFoundError

__all__ = [
    "require_s3",
    "require_toml",
    "require_xls",
    "require_xml",
    "require_yaml",
]


def _require_optional_module(*, target, installed):
    if not installed:
        raise ExtrasRequireModuleNotFoundError(target=target)


def require_s3(*, installed):
    _require_optional_module(target="s3", installed=installed)


def require_toml(*, installed):
    _require_optional_module(target="toml", installed=installed)


def require_xls(*, installed):
    _require_optional_module(target="xls", installed=installed)


def require_xml(*, installed):
    _require_optional_module(target="xml", installed=installed)


def require_yaml(*, installed):
    _require_optional_module(target="yaml", installed=installed)



class ExtrasRequireModuleNotFoundError(ModuleNotFoundError):
    def __init__(self, *, target):
        message = (
            f"Extras require '[{target}]' module not found, "
            f"you can install it by running: "
            f"'python -m pip install \"python-benedict[{target}]\"'"
        )
        super().__init__(message)



import re
from datetime import datetime
from decimal import Decimal, DecimalException

import ftfy
import phonenumbers
from dateutil import parser as date_parser
from MailChecker import MailChecker
from phonenumbers import PhoneNumberFormat, phonenumberutil
from slugify import slugify

from benedict.serializers import JSONSerializer
from benedict.utils import type_util


def _parse_with(val, type_checker, parser, **kwargs):
    if val is None:
        return None
    if callable(type_checker) and type_checker(val):
        return val
    s = str(val)
    if not len(s):
        return None
    return parser(s, **kwargs)


def _parse_bool(val):
    val = val.lower()
    if val in ["1", "true", "yes", "ok", "on"]:
        return True
    elif val in ["0", "false", "no", "ko", "off"]:
        return False
    return None


def parse_bool(val):
    return _parse_with(val, type_util.is_bool, _parse_bool)


def parse_date(val, format=None):
    val = parse_datetime(val, format)
    if val:
        return val.date()
    return None


def _parse_datetime_with_format(val, format):
    try:
        return datetime.strptime(val, format)
    except Exception:
        return None


def _parse_datetime_without_format(val):
    try:
        return date_parser.parse(val)
    except Exception:
        return _parse_datetime_from_timestamp(val)


def _parse_datetime_from_timestamp(val):
    try:
        return datetime.fromtimestamp(float(val))
    except Exception:
        return None


def parse_datetime(val, format=None):
    if type_util.is_datetime(val):
        return val
    s = str(val)
    if format:
        return _parse_datetime_with_format(s, format)
    else:
        return _parse_datetime_without_format(s)


def _parse_decimal(val):
    try:
        return Decimal(val)
    except (ValueError, DecimalException):
        return None


def parse_decimal(val):
    return _parse_with(val, type_util.is_decimal, _parse_decimal)


def _parse_dict(val):
    serializer = JSONSerializer()
    try:
        d = serializer.decode(val)
        if type_util.is_dict(d):
            return d
        return None
    except Exception:
        return None


def parse_dict(val):
    return _parse_with(val, type_util.is_dict, _parse_dict)


def _parse_float(val):
    try:
        return float(val)
    except ValueError:
        return None


def parse_float(val):
    return _parse_with(val, type_util.is_float, _parse_float)


def _parse_email(val, check_blacklist=True):
    val = val.lower()
    if check_blacklist:
        if not MailChecker.is_valid(val):
            return None
    else:
        if not MailChecker.is_valid_email_format(val):
            return None
    return val


def parse_email(val, check_blacklist=True):
    return _parse_with(val, None, _parse_email, check_blacklist=check_blacklist)


def _parse_int(val):
    try:
        return int(val)
    except ValueError:
        return None


def parse_int(val):
    return _parse_with(val, type_util.is_integer, _parse_int)


def _parse_list(val, separator=None):
    if (
        val.startswith("{")
        and val.endswith("}")
        or val.startswith("[")
        and val.endswith("]")
    ):
        try:
            serializer = JSONSerializer()
            ls = serializer.decode(val)
            if type_util.is_list(ls):
                return ls
            return None
        except Exception:
            pass
    if separator:
        ls = list(val.split(separator))
        return ls
    return None


def parse_list(val, separator=None):
    val = _parse_with(val, type_util.is_list_or_tuple, _parse_list, separator=separator)
    return list(val) if type_util.is_list_or_tuple(val) else val


def _parse_phonenumber(val, country_code=None):
    try:
        phone_obj = phonenumbers.parse(val, country_code)
        if phonenumbers.is_valid_number(phone_obj):
            return {
                "e164": phonenumbers.format_number(phone_obj, PhoneNumberFormat.E164),
                "international": phonenumbers.format_number(
                    phone_obj, PhoneNumberFormat.INTERNATIONAL
                ),
                "national": phonenumbers.format_number(
                    phone_obj, PhoneNumberFormat.NATIONAL
                ),
            }
        return None
    except phonenumberutil.NumberParseException:
        return None


def parse_phonenumber(val, country_code=None):
    s = parse_str(val)
    if not s:
        return None
    phone_raw = re.sub(r"[^0-9\+]", " ", s)
    phone_raw = phone_raw.strip()
    if phone_raw.startswith("00"):
        phone_raw = "+" + phone_raw[2:]
    if country_code and len(country_code) >= 2:
        country_code = country_code[0:2].upper()
    return _parse_with(phone_raw, None, _parse_phonenumber, country_code=country_code)


def _parse_slug(val):
    return slugify(val)


def parse_slug(val):
    s = parse_str(val)
    return _parse_slug(s)


def parse_str(val):
    if type_util.is_string(val):
        val = ftfy.fix_text(val)
    else:
        val = str(val)
    val = val.strip()
    val = " ".join(val.split())
    return val


def parse_uuid(val):
    s = parse_str(val)
    return s if type_util.is_uuid(s) else None



from benedict.dicts.base import BaseDict


class KeyattrDict(BaseDict):
    _keyattr_enabled = None

    def __init__(self, *args, **kwargs):
        self._keyattr_enabled = kwargs.pop("keyattr_enabled", True)
        super().__init__(*args, **kwargs)

    @property
    def keyattr_enabled(self):
        return self._keyattr_enabled

    @keyattr_enabled.setter
    def keyattr_enabled(self, value):
        self._keyattr_enabled = value

    def __getattr__(self, attr):
        if not self._keyattr_enabled:
            raise AttributeError
        # return super().__getattr__(attr)
        try:
            return self.__getitem__(attr)
        except KeyError:
            self.__setitem__(attr, {})
            return self.__getitem__(attr)

    def __setattr__(self, attr, value):
        if attr in self:
            # set existing key
            if not self._keyattr_enabled:
                raise AttributeError
            self.__setitem__(attr, value)
        elif hasattr(self.__class__, attr):
            # set existing attr
            super().__setattr__(attr, value)
        else:
            # set new key
            if not self._keyattr_enabled:
                raise AttributeError
            self.__setitem__(attr, value)

    def __setstate__(self, state):
        super().__setstate__(state)
        self._keyattr_enabled = state["_keyattr_enabled"]



__author__ = "Fabio Caccamo"
__copyright__ = "Copyright (c) 2019-present - Fabio Caccamo"
__description__ = (
    "python-benedict is a dict subclass with keylist/keypath support, normalized I/O"
    " operations (base64, csv, ini, json, pickle, plist, query-string, toml, xls, xml, yaml)"
    " and many utilities... for humans, obviously."
)
__email__ = "fabio.caccamo@gmail.com"
__license__ = "MIT"
__title__ = "benedict"
__version__ = "0.29.0"



from benedict.core.groupby import groupby


def _nest_items(nested_items, item, id_key, children_key):
    children_items = nested_items.pop(item[id_key], [])
    item[children_key] = children_items
    for child_item in children_items:
        _nest_items(nested_items, child_item, id_key, children_key)


def nest(items, id_key, parent_id_key, children_key):
    if any(
        [id_key == parent_id_key, id_key == children_key, parent_id_key == children_key]
    ):
        raise ValueError("keys should be different.")
    nested_items = groupby(items, parent_id_key)
    root_items = nested_items.get(None, [])
    for item in root_items:
        _nest_items(nested_items, item, id_key, children_key)
    return nested_items.get(None)



try:
    import yaml

    yaml_installed = True
except ModuleNotFoundError:
    yaml_installed = False


from benedict.extras import require_yaml
from benedict.serializers.abstract import AbstractSerializer
from benedict.serializers.json import JSONSerializer


class YAMLSerializer(AbstractSerializer):
    """
    This class describes an yaml serializer.
    """

    def __init__(self):
        super().__init__(
            extensions=[
                "yaml",
                "yml",
            ],
        )
        self._json_serializer = JSONSerializer()

    def decode(self, s, **kwargs):
        require_yaml(installed=yaml_installed)
        data = yaml.safe_load(s, **kwargs)
        return data

    def encode(self, d, **kwargs):
        require_yaml(installed=yaml_installed)
        d = self._json_serializer.decode(self._json_serializer.encode(d))
        data = yaml.dump(d, **kwargs)
        return data



import fsutil

try:
    from openpyxl import load_workbook
    from xlrd import open_workbook

    xls_installed = True
except ModuleNotFoundError:
    xls_installed = False

from slugify import slugify

from benedict.extras import require_xls
from benedict.serializers.abstract import AbstractSerializer


class XLSSerializer(AbstractSerializer):
    """
    This class describes a xls serializer.
    """

    def __init__(self):
        super().__init__(
            extensions=[
                "xls",
                "xlsx",
                "xlsm",
            ],
        )

    def _get_sheet_index_and_name_from_options(self, **kwargs):
        sheet_index_or_name = kwargs.pop("sheet", 0)
        sheet_index = 0
        sheet_name = ""
        if isinstance(sheet_index_or_name, int):
            sheet_index = sheet_index_or_name
        elif isinstance(sheet_index_or_name, str):
            sheet_name = sheet_index_or_name
        return (sheet_index, sheet_name)

    def _get_sheet_index_by_name(self, sheet_name, sheet_names):
        sheet_names = list([slugify(name) for name in sheet_names])
        try:
            sheet_index = sheet_names.index(slugify(sheet_name))
            return sheet_index
        except ValueError as error:
            raise Exception(
                f"Invalid sheet name {sheet_name!r}, sheet not found."
            ) from error

    def _get_sheet_columns_indexes(self, columns_count):
        return [column_index for column_index in range(columns_count)]

    def _decode_legacy(self, s, **kwargs):
        filepath = s

        # load the worksheet
        workbook = open_workbook(filename=filepath)

        # get sheet by index or by name
        sheet_index, sheet_name = self._get_sheet_index_and_name_from_options(**kwargs)
        if sheet_name:
            sheet_names = workbook.sheet_names()
            sheet_index = self._get_sheet_index_by_name(sheet_name, sheet_names)
        sheet = workbook.sheet_by_index(sheet_index)
        sheet_columns_range = range(sheet.ncols)

        # get columns
        columns = kwargs.pop("columns", None)
        columns_row = kwargs.pop("columns_row", True)
        columns_standardized = kwargs.pop("columns_standardized", columns is None)
        if not columns:
            if columns_row:
                # if first row is for column names read the names
                # for row in sheet.iter_rows(min_row=1, max_row=1):
                columns = [
                    sheet.cell_value(0, col_index) for col_index in sheet_columns_range
                ]
            else:
                # otherwise use columns indexes as column names
                # for row in sheet.iter_rows(min_row=1, max_row=1):
                columns = self._get_sheet_columns_indexes(sheet_columns_range)

        # standardize column names, eg. "Date Created" -> "date_created"
        if columns_standardized:
            columns = [slugify(column, separator="_") for column in columns]

        # build list of dicts, one for each row
        items = []
        items_row_start = 1 if columns_row else 0
        for row_index in range(items_row_start, sheet.nrows):
            row = {}
            for col_index in sheet_columns_range:
                col_key = columns[col_index]
                value = sheet.cell_value(row_index, col_index)
                row[col_key] = value
            items.append(row)

        # print(items)
        return items

    def _decode(self, s, **kwargs):
        filepath = s

        # load the worksheet
        workbook = load_workbook(filename=filepath, read_only=True)

        # get sheet by index or by name
        sheet_index, sheet_name = self._get_sheet_index_and_name_from_options(**kwargs)
        sheets = [sheet for sheet in workbook]
        if sheet_name:
            sheet_names = [sheet.title for sheet in sheets]
            sheet_index = self._get_sheet_index_by_name(sheet_name, sheet_names)
        sheet = sheets[sheet_index]
        sheet_columns_cells = list(sheet.iter_rows(min_row=1, max_row=1))[0]

        # get columns
        columns = kwargs.pop("columns", None)
        columns_row = kwargs.pop("columns_row", True)
        columns_standardized = kwargs.pop("columns_standardized", columns is None)
        if not columns:
            if columns_row:
                # if first row is for column names read the names
                # for row in sheet.iter_rows(min_row=1, max_row=1):
                columns = [cell.value for cell in sheet_columns_cells]
            else:
                # otherwise use columns indexes as column names
                # for row in sheet.iter_rows(min_row=1, max_row=1):
                columns = self._get_sheet_columns_indexes(len(sheet_columns_cells))

        # standardize column names, eg. "Date Created" -> "date_created"
        if columns_standardized:
            columns = [slugify(column, separator="_") for column in columns]

        # build list of dicts, one for each row
        items = []
        items_row_start = 2 if columns_row else 1
        for row in sheet.iter_rows(min_row=items_row_start):
            values = list([cell.value for cell in row])
            items.append(dict(zip(columns, values)))

        # close the worksheet
        workbook.close()

        # print(items)
        return items

    def decode(self, s, **kwargs):
        require_xls(installed=xls_installed)
        extension = fsutil.get_file_extension(s)
        if extension in ["xlsx", "xlsm"]:
            return self._decode(s, **kwargs)
        elif extension in ["xls", "xlt"]:
            return self._decode_legacy(s, **kwargs)

    def encode(self, d, **kwargs):
        # require_xls(installed=xls_installed)
        raise NotImplementedError



import re

from benedict.serializers.abstract import AbstractSerializer
from benedict.serializers.base64 import Base64Serializer
from benedict.serializers.csv import CSVSerializer
from benedict.serializers.ini import INISerializer
from benedict.serializers.json import JSONSerializer
from benedict.serializers.pickle import PickleSerializer
from benedict.serializers.plist import PListSerializer
from benedict.serializers.query_string import QueryStringSerializer
from benedict.serializers.toml import TOMLSerializer
from benedict.serializers.xls import XLSSerializer
from benedict.serializers.xml import XMLSerializer
from benedict.serializers.yaml import YAMLSerializer

__all__ = [
    "AbstractSerializer",
    "Base64Serializer",
    "CSVSerializer",
    "INISerializer",
    "JSONSerializer",
    "PickleSerializer",
    "PListSerializer",
    "QueryStringSerializer",
    "TOMLSerializer",
    "XLSSerializer",
    "XMLSerializer",
    "YAMLSerializer",
]

_BASE64_SERIALIZER = Base64Serializer()
_CSV_SERIALIZER = CSVSerializer()
_INI_SERIALIZER = INISerializer()
_JSON_SERIALIZER = JSONSerializer()
_PICKLE_SERIALIZER = PickleSerializer()
_PLIST_SERIALIZER = PListSerializer()
_QUERY_STRING_SERIALIZER = QueryStringSerializer()
_TOML_SERIALIZER = TOMLSerializer()
_XLS_SERIALIZER = XLSSerializer()
_XML_SERIALIZER = XMLSerializer()
_YAML_SERIALIZER = YAMLSerializer()

_SERIALIZERS_LIST = [
    _BASE64_SERIALIZER,
    _CSV_SERIALIZER,
    _INI_SERIALIZER,
    _JSON_SERIALIZER,
    _PICKLE_SERIALIZER,
    _PLIST_SERIALIZER,
    _QUERY_STRING_SERIALIZER,
    _TOML_SERIALIZER,
    _YAML_SERIALIZER,
    _XLS_SERIALIZER,
    _XML_SERIALIZER,
]

_SERIALIZERS_BY_EXTENSION = {}
for serializer in _SERIALIZERS_LIST:
    for extension in serializer.extensions():
        _SERIALIZERS_BY_EXTENSION[extension] = serializer

_SERIALIZERS_EXTENSIONS = [
    f".{extension}" for extension in _SERIALIZERS_BY_EXTENSION.keys()
]


def get_format_by_path(path):
    path = str(path)
    path = path.lower()
    for extension in _SERIALIZERS_EXTENSIONS:
        if path.endswith(extension):
            return extension[1:]
    return None


def get_serializer_by_format(format):
    format_key = (format or "").lower().strip()
    format_key = re.sub(r"[\s\-\_]*", "", format_key)
    serializer = _SERIALIZERS_BY_EXTENSION.get(format_key, None)
    return serializer



import json

from benedict.serializers.abstract import AbstractSerializer
from benedict.utils import type_util


class JSONSerializer(AbstractSerializer):
    """
    This class describes a json serializer.
    """

    def __init__(self):
        super().__init__(
            extensions=[
                "json",
            ],
        )

    def decode(self, s, **kwargs):
        data = json.loads(s, **kwargs)
        return data

    def encode(self, d, **kwargs):
        kwargs.setdefault("default", self._encode_default)
        data = json.dumps(d, **kwargs)
        return data

    def _encode_default(self, obj):
        if type_util.is_set(obj):
            return list(obj)
        elif type_util.is_datetime(obj):
            return obj.isoformat()
        elif type_util.is_decimal(obj):
            return str(obj)
        return str(obj)



try:
    import toml

    toml_installed = True
except ModuleNotFoundError:
    toml_installed = False

try:
    # python >= 3.11
    import tomllib

    tomllib_available = True
except ImportError:
    tomllib_available = False

from benedict.extras import require_toml
from benedict.serializers.abstract import AbstractSerializer


class TOMLSerializer(AbstractSerializer):
    """
    This class describes a toml serializer.
    """

    def __init__(self):
        super().__init__(
            extensions=[
                "toml",
            ],
        )

    def decode(self, s, **kwargs):
        if tomllib_available:
            data = tomllib.loads(s, **kwargs)
        else:
            require_toml(installed=toml_installed)
            data = toml.loads(s, **kwargs)
        return data

    def encode(self, d, **kwargs):
        require_toml(installed=toml_installed)
        data = toml.dumps(dict(d), **kwargs)
        return data



import pathlib
import re
from datetime import datetime
from decimal import Decimal

regex = re.compile("").__class__
uuid_re = re.compile(
    "^([0-9a-f]{32}){1}$|^([0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}){1}$",
    flags=re.IGNORECASE,
)


def is_bool(val):
    return isinstance(val, bool)


def is_collection(val):
    return isinstance(val, (dict, list, set, tuple))


def is_datetime(val):
    return isinstance(val, datetime)


def is_decimal(val):
    return isinstance(val, Decimal)


def is_dict(val):
    return isinstance(val, dict)


def is_dict_or_list(val):
    return isinstance(val, (dict, list))


def is_dict_or_list_or_tuple(val):
    return isinstance(val, (dict, list, tuple))


def is_float(val):
    return isinstance(val, float)


def is_function(val):
    return callable(val)


def is_integer(val):
    return isinstance(val, int)


def is_json_serializable(val):
    json_types = (type(None), bool, dict, float, int, list, str, tuple)
    return isinstance(val, json_types)


def is_list(val):
    return isinstance(val, list)


def is_list_or_tuple(val):
    return isinstance(val, (list, tuple))


def is_none(val):
    return val is None


def is_not_none(val):
    return val is not None


def is_path(val):
    return isinstance(val, pathlib.Path)


def is_regex(val):
    return isinstance(val, regex)


def is_set(val):
    return isinstance(val, set)


def is_string(val):
    return isinstance(val, str)


def is_tuple(val):
    return isinstance(val, tuple)


def is_uuid(val):
    return is_string(val) and uuid_re.match(val)
